{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards #-}

module TiedoteMD.Read where

import Control.Exception (throwIO)
import Control.Monad ((>=>), join, unless)
import Control.Monad.IO.Class (liftIO)
import Crypto.Hash (hashWith, SHA256(..))
import Data.Acid (AcidState, update)
import Data.Bifunctor (first, second)
import Data.ByteArray (convert)
import Data.ByteString (ByteString)
import Data.Default (def)
import Data.Either (rights, lefts)
import Data.FileStore (FileStore(..), gitFileStore)
import Data.List (singleton)
import Data.Text.Encoding (decodeUtf8, encodeUtf8)
import Data.Time (UTCTime, zonedTimeToUTC)
import Data.Time (getCurrentTime)
import Data.Time.Format.ISO8601 (iso8601ParseM)
import System.Exit (ExitCode(..))
import System.IO (hClose)
import System.Process (createProcess, proc, waitForProcess, CreateProcess(..), StdStream(..))
import Text.Pandoc (Pandoc(..), ReaderOptions(..), WriterOptions(..), Meta, MetaValue(..), nullMeta, Inline, Block(Plain), runIO, lookupMeta, runPure, pandocExtensions)
import Text.Pandoc.Readers (readMarkdown)
import Text.Pandoc.Writers (writePlain, writeMarkdown, writeHtml5String)

import qualified Data.ByteString as BS
import qualified Data.Text as T
import qualified Data.Text.IO as T

import TiedoteMD.Git
import TiedoteMD.State
import TiedoteMD.Templates
import TiedoteMD.Types

updateMessages :: AcidState State -> FilePath -> IO ()
updateMessages acid repoPath =
    gitPull repoPath >> readMessageFiles (gitFileStore repoPath) >>= update acid . UpdateState

readMessageFiles :: FileStore -> IO [Message]
readMessageFiles store = do
    files <- index store 
    messages <- mapM (readMessageFile store) files
    mapM_ (T.putStrLn . renderError) $ lefts messages
    currentTime <- getCurrentTime
    print currentTime
    pure $ filter ((currentTime <) . sendTime) $ rights messages
-- TODO: files could be read in parallel instead of sequentially with mapM

readMessageFile :: FileStore -> FilePath -> IO (Either Error Message)
readMessageFile store = flip (retrieve store) Nothing >=>
    uncurry parseMessageFile . addHash . T.filter (/= '\r') . decodeUtf8 . BS.toStrict
        where addHash text = (convert $ hashWith SHA256 $ encodeUtf8 text, text)

parseMessageFile :: ByteString -> T.Text -> IO (Either Error Message)
parseMessageFile hash text = fmap (join . first PandocError) . runIO $ do
    pandoc@(Pandoc meta _) <- flip readMarkdown text
        def {readerStandalone = True, readerExtensions = pandocExtensions}
    plainTextMessage <- renderHelper writePlain plainTemplate pandoc
    markdownMessage <- renderHelper writeMarkdown markdownTemplate pandoc
    htmlMessage <- liftIO . inlineCSS =<< renderHelper writeHtml5String htmlTemplate pandoc
    pure $ do
        previewTo <- lookupMeta' "tarkistaja" meta >>= metaToEmails
        previewTime <- lookupMeta' "deadline" meta >>= metaToTime
        sendTime <- lookupMeta' "lähetysaika" meta >>= metaToTime
        subject <- lookupMeta' "otsikko" meta >>= metaToText
        pure Message
            { recipients = []
            , messageHash = hash
            , messageContent = MessageContent {..}
            , previewMailID = Nothing
            , ..
            }
        where renderHelper writer template = writer (writerOpts {writerTemplate = Just template})
              writerOpts = def
                  { writerTOCDepth = 2
                  , writerTableOfContents = True
                  , writerSectionDivs = True
                  }
    -- TODO: Store the media somewhere

inlineCSS :: T.Text -> IO T.Text
inlineCSS html = do
    (Just hIn, Just hOut, _, processHandle) <- createProcess $
        (proc "douceur" ["inline", "/dev/stdin"]) {std_in = CreatePipe, std_out = CreatePipe}
    T.hPutStrLn hIn html
    hClose hIn
    inlined <- T.hGetContents hOut
    exitCode <- waitForProcess processHandle
    unless (exitCode == ExitSuccess) $ throwIO $ ProcessError "douceur" exitCode
    pure inlined

metaToTime :: MetaValue -> Either Error UTCTime
metaToTime meta = do
    textMeta <- metaToTexts meta >>= headOrError
    maybe (Left $ InvalidTime textMeta) pure $
        fmap zonedTimeToUTC $ iso8601ParseM $ T.unpack $ T.strip textMeta
            where headOrError (x:_) = pure x
                  headOrError _ = Left $ InvalidTime ""

metaToEmails :: MetaValue -> Either Error [Email]
metaToEmails meta = metaToTexts meta >>= mapM
    (\text -> maybe (Left $ InvalidEmail text) pure $ email $ encodeUtf8 $ T.strip text)

metaToText :: MetaValue -> Either Error T.Text
metaToText = second T.unwords . metaToTexts

metaToTexts :: MetaValue -> Either Error [T.Text]
metaToTexts (MetaString text) = pure [text]
metaToTexts (MetaInlines inlines) = second singleton $ inlinesToText inlines
metaToTexts (MetaList metas) = second concat $ mapM metaToTexts metas
metaToTexts _ = Left InvalidData

inlinesToText :: [Inline] -> Either Error T.Text
inlinesToText = first PandocError . runPure . writeMarkdown def{writerExtensions = pandocExtensions} . Pandoc nullMeta . singleton . Plain

lookupMeta' :: T.Text -> Meta -> Either Error MetaValue
lookupMeta' key meta = maybe (Left $ NoMeta key) Right $ lookupMeta key meta
